#! /usr/bin/env python3

"""Main XOInvader module, that is entry point to game.

Prepare environment for starting game and start it."""


import argparse
import logging

import xoinvader
from xoinvader.app import XOInvader


LOG = logging.getLogger(__name__)


def create_game():
    """Create XOInvader game instance."""

    app = xo1.Application(
        Settings.layout.field.border.x, Settings.layout.field.border.y,
    )

    return app


def parse_args():
    """Parse incoming arguments."""
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-d", "--debug", action="store_true", help="enable debug mode"
    )

    args = parser.parse_args()
    return args


def main():
    """Start the game!"""

    args = parse_args()
    # apply settings to engine settings
    xoinvader.init(args.__dict__)
    LOG.debug("Incoming args: %s", args)
    game = XOInvader()
    return game.start()


if __name__ == "__main__":
    main()



"""XOInvader styles."""

import curses

from eaf.meta import Singleton


class Style(metaclass=Singleton):
    """Container for style mappings."""

    def __init__(self):
        self._style = {
            "gui": {},
            "obj": {},
        }

    def init_styles(self, palette):
        """Initialize styles."""

        cpair = curses.color_pair

        self.gui["normal"] = cpair(palette.ui_norm) | curses.A_BOLD
        self.gui["yellow"] = cpair(palette.ui_yellow) | curses.A_BOLD
        self.gui["dp_blank"] = cpair(palette.dp_blank) | curses.A_BOLD
        self.gui["dp_ok"] = cpair(palette.dp_ok) | curses.A_BOLD
        self.gui["dp_middle"] = cpair(palette.dp_middle) | curses.A_BOLD
        self.gui["dp_critical"] = cpair(palette.dp_critical) | curses.A_BOLD
        self.gui["sh_ok"] = cpair(palette.sh_ok) | curses.A_BOLD
        self.gui["sh_mid"] = cpair(palette.sh_mid) | curses.A_BOLD

    def __getattr__(self, name):
        return self._style[name]



"""Base handler class."""

from abc import ABCMeta, abstractmethod


class Handler(metaclass=ABCMeta):
    """Base game handler.

    Provides accessing to main State's fields and handle function stub.

    :param :class:`xoinvader.state.State` owner: handler's owner
    """

    def __init__(self, owner):
        self._owner = owner

    @property
    def owner(self):
        """Handler's owner.

        :getter: yes
        :setter: no
        :type: :class:`xoinvader.state.State`
        """

        return self._owner

    @abstractmethod
    def handle(self):
        """Handle event."""

        pass  # pragma: no cover


# TODO: event-handling: now this is prototype that must be reworked in future
# TODO: remove owner link if it's not needed anymore
class EventHandler(Handler):
    """Base game event handler.

    Handles events using the event queue.
    Provides command mapping manipulations via the default handle() method.
    You must instantiate this class only after application being initialized.
    Application instance has event queue object, that will be used here.

    :param :class:`xoinvader.state.State` owner: handler's owner state
    :param dict command_map: key->command mapping
    """

    def __init__(self, owner, command_map=None):
        super(EventHandler, self).__init__(owner)

        self._command_map = command_map or {}
        # TODO: implement proper queue object
        self._event_queue = owner.app.event_queue

    # TODO: event-handling: I think we need some global event bus in whole
    #       application.
    def event_queue(self):
        """Event queue stub."""

        return [("KEY_PRESS", self.get_input())]

    # TODO: event-handling: abstract getting input from curses
    def get_input(self):
        """Get input from keyboard."""

        return self._event_queue.getch()

    def handle(self):
        """Handle input event."""

        for event in self.event_queue():
            if event[0] == "KEY_PRESS":
                command = self._command_map.get(event[1])
                if callable(command):
                    command()
            else:
                raise ValueError("Unknown event type: {0}".format(event[0]))



"""MainMenuState-related input and event handlers."""

from eaf.state import State

from xoinvader import app
from xoinvader.gui import (
    MenuItemContainer,
    MenuItemWidget,
    PopUpNotificationWidget,
    TextCallbackWidget,
    TextWidget,
)
from xoinvader.handlers import EventHandler
from xoinvader.keys import KEY
from xoinvader.utils import Point


class PauseMenuState(State):
    def __init__(self, app):
        super().__init__(app)

        self.add(TextWidget(Point(4, 4), "Pause"))
        self._items = MenuItemContainer(
            [
                MenuItemWidget(
                    Point(10, 10),
                    "Continue",
                    template=("=> ", ""),
                    action=lambda: app.current().trigger_state("InGameState"),
                ),
                MenuItemWidget(
                    Point(10, 11),
                    "Quit",
                    template=("=> ", ""),
                    action=app.current().stop,
                ),
            ]
        )
        self._items.select(0)
        self.add(self._items)

        self._events = EventHandler(
            self,
            {
                KEY.ESCAPE: lambda: app.current().trigger_state("InGameState"),
                KEY.SPACE: self._items.do_action,
                KEY.W: self._items.prev,
                KEY.S: self._items.next,
                KEY.N: lambda: self.notify("This is test notification"),
            },
        )

    def notify(self, text, pos=Point(15, 15)):
        self.add(
            PopUpNotificationWidget(
                pos, text, callback=lambda w: self.remove(w)
            )
        )

    def events(self):
        self._events.handle()


class GameOverState(State):
    def __init__(self, app):
        super().__init__(app)

        self._score = "Your score: {0}"

        self.add(
            TextWidget(Point(4, 4), "I was too tired and scared to continue...")
        )
        self.add(TextCallbackWidget(Point(4, 5), self.score_callback))

        self._items = MenuItemContainer(
            [
                MenuItemWidget(
                    Point(10, 10),
                    "I must to continue my revenge!",
                    action=lambda: app.current().trigger_reinit("InGameState"),
                ),
                MenuItemWidget(
                    Point(10, 11),
                    "No! I want to go home!",
                    action=app.current().stop,
                ),
            ]
        )
        self.add(self._items)
        self._items.select(0)

        self._events = EventHandler(
            self,
            {
                KEY.SPACE: self._items.do_action,
                KEY.W: self._items.prev,
                KEY.S: self._items.next,
            },
        )

    def score_callback(self):
        return self._score

    def trigger(self, score):
        """Trigger the state and pass the score info to it.

        :param int score: last player score
        """

        self._score = self._score.format(score)

    def events(self):
        self._events.handle()



"""XOInvader game application class."""

from xoinvader import Settings
from xoinvader.ingame import InGameState
from xoinvader.menu import PauseMenuState, GameOverState
from xoinvader.style import Style

from xo1 import Application, Palette


class XOInvader(Application):
    """XOInvader game application class."""

    def __init__(self):

        palette = Palette(
            [
                # User interface colors
                ("ui_norm", Palette.COLOR_WHITE, Palette.COLOR_BLACK),
                ("ui_yellow", Palette.COLOR_YELLOW, Palette.COLOR_BLACK),
                # Damage panel colors
                ("dp_blank", Palette.COLOR_BLACK, Palette.COLOR_BLACK),
                ("dp_ok", Palette.COLOR_GREEN, Palette.COLOR_BLACK),
                ("dp_middle", Palette.COLOR_YELLOW, Palette.COLOR_BLACK),
                ("dp_critical", Palette.COLOR_RED, Palette.COLOR_BLACK),
                ("sh_ok", Palette.COLOR_BLUE, Palette.COLOR_BLACK),
                ("sh_mid", Palette.COLOR_CYAN, Palette.COLOR_BLACK),
                # Weapon's charge colors
                ("blaster", Palette.COLOR_GREEN, Palette.COLOR_BLACK),
                ("laser", Palette.COLOR_BLACK, Palette.COLOR_RED),
                ("um", Palette.COLOR_MAGENTA, Palette.COLOR_BLACK),
            ]
        )

        super().__init__(
            x=Settings.layout.field.border.x,
            y=Settings.layout.field.border.y,
            palette=palette,
        )

        Style().init_styles(palette)

        self.register(InGameState)
        self.register(PauseMenuState)
        self.register(GameOverState)


def current():
    return XOInvader.current()



"""Pytest configuration."""

from unittest import mock

import pytest

from eaf.app
from eaf.state import State

import xoinvader.app


@pytest.fixture
def mock_application(request):

    app = None

    def inner():
        nonlocal app

        app = MockedApplication()
        return app

    request.addfinalizer(MockedApplication._finalize)
    return inner


@pytest.fixture
def mock_state(request, mock_application):

    app = None

    def inner(mock_app=False):
        nonlocal app

        if mock_app:
            # We need to create reference or object will be collected by gc
            app = mock_application()
        else:
            app = xoinvader.app.current()

        app.register(MockedState)
        return app.state

    def stop():
        app.deregister(MockedState.__name__)

    request.addfinalizer(stop)

    return inner


class MockedApplication(xoinvader.app.XOInvader):

    @staticmethod
    def _finalize():
        try:
            app = xoinvader.app.current()
            app.stop()
            del app
        except:
            pass


class MockedState(State):
    pass



"""Test xoinvader.handlers module."""

import pytest

from xoinvader import handlers


class EventQueueMock:
    """Event queue mock."""

    key = "test-key"

    def getch(self):
        """Mocked method to emit pressed key."""

        return self.key

    def mocked_event_queue(self):  # pylint: disable=no-self-use
        """Mocked method to emit wrong event."""

        return [("bad-event", "test")]


class OwnerMock:  # pylint: disable=too-few-public-methods
    """Owner mock to provide access to app and event queue for handler."""

    @property
    def app(cls):
        return cls

    @property
    def event_queue(cls):
        return EventQueueMock()


def test_event_handler():
    """Test xoinvader.handlers.EventHandler."""

    events = handlers.EventHandler(OwnerMock())

    assert isinstance(events.owner, OwnerMock)
    assert isinstance(events._event_queue, EventQueueMock)

    assert events.get_input() == EventQueueMock.key
    assert events.event_queue() == [("KEY_PRESS", EventQueueMock.key)]

    assert not events.handle()

    fired = []
    events = handlers.EventHandler(
        OwnerMock(),
        {EventQueueMock.key: lambda: fired.append(EventQueueMock.key)},
    )

    events.handle()
    assert fired == [EventQueueMock.key]


def test_event_handler_negative(monkeypatch):
    """Negatively test xoinvader.handlers.EventHandler."""

    monkeypatch.setattr(
        handlers.EventHandler, "event_queue", EventQueueMock.mocked_event_queue
    )

    events = handlers.EventHandler(OwnerMock())
    with pytest.raises(ValueError):
        events.handle()



"""Enemy and player ships."""

import logging

from xoinvader import app
from xoinvader import collision
from xoinvader.animation import AnimationManager
from xoinvader.render import Renderable
from xoinvader.weapon import Blaster, Laser, UM, EBlaster
from xoinvader.utils import Point, Surface, InfiniteList
from xoinvader.collision import Collider
from xoinvader.common import Settings, get_config


LOG = logging.getLogger(__name__)

# pylint: disable=missing-docstring
CONFIG = get_config().ship


# Think about composition
class Ship(Renderable):
    """Base class for all ships. Contains basic ship logic."""

    compound = True

    def __init__(self, pos):
        self._type = self.__class__.__name__
        self._image = None

        self._pos = pos

        self._dx = None
        self._fire = False
        self._weapon = None
        self._weapons = InfiniteList()
        self._wbay = None
        self._direction = 0

        self._max_hull = None
        self._max_shield = None
        self._hull = None
        self._shield = None

        # Set to True means that object in destroying phase
        # and will ignore remove_obsolete signal
        self._destroy = False

        self._collider = None

        # first initialization
        self._load_config(CONFIG[self._type])

    def _load_config(self, config):
        """Load config from mapping."""

        if not config:
            raise ValueError

        self._dx = int(config.dx)
        self._hull = int(config.hull)
        self._shield = int(config.shield)
        self._max_hull = int(config.max_hull)
        self._max_shield = int(config.max_shield)

    @property
    def pos(self):
        return self._pos

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, value):
        if value == 0:
            self._direction = 0
        else:
            self._direction = 1 if value > 0 else -1

    @property
    def max_hull(self):
        return self._max_hull

    @property
    def max_shield(self):
        return self._max_shield

    def get_hull_percentage(self):
        """Return hull percentage."""
        return self._hull * 100.0 / self._max_hull

    def get_shield_percentage(self):
        """Return shield percentage."""
        return self._shield * 100.0 / self._max_shield

    def get_weapon_percentage(self):
        """Return weapon load percentage."""
        return self._weapon.load_percentage()

    def get_render_data(self):
        """Callback for rendering."""
        return [self._pos], self._image.get_image()

    def get_renderable_objects(self):
        """CORP stub."""
        return self._weapons

    def move_left(self):
        """Change direction."""
        self._direction = -1

    def move_right(self):
        """Change direction."""
        self._direction = 1

    def toggle_fire(self):
        """Toggle current weapon fire mode."""
        self._fire = not self._fire

    def next_weapon(self):
        """Select next weapon."""
        self._weapon = self._weapons.next()

    def prev_weapon(self):
        """Select previous weapon."""
        self._weapon = self._weapons.prev()

    def add_weapon(self, weapon):
        """Add new weapon."""

        self._weapons.append(weapon)
        self._weapon = weapon

    def update_position(self, dt):
        """Update ship position.

        Allows to go behind field borders.
        """

        self._pos.x += self._direction * self._dx * dt
        self._direction = 0

    def update(self, dt):
        """Update ship object's state."""

        self.update_position(dt)

        for weapon in self._weapons:
            weapon.update(dt)

        if self._fire:
            try:
                # FIXME: [fix-in-place-coercing]
                #        Coercing Point in-place smells badly
                #        (needed now because animation can set float pos to
                #        enemy ships), we need to add type ensurance mechanism
                #        to cast automatically maybe.
                self._weapon.make_shot(self._pos[int] + self._wbay)
            except ValueError:
                self.next_weapon()

        self.refresh_shield()

    def destroy(self):
        """Self-destroying routine."""

        LOG.debug("Destroying ship %s", self)

        if not self._destroy:
            self._destroy = True
            app.current().state.collision.remove(self._collider)
            app.current().state.remove(self)

    def remove_obsolete(self, pos):
        border = Settings.layout.field.border
        pos = self._pos[int]
        if (
            pos.x > self._image.width + border.x
            or pos.x + self._image.width < 0
            or int(pos.y) > self._image.height + border.y
            or int(pos.y) + self._image.height < 0
        ) and not self._destroy:
            self.destroy()

    def take_damage(self, damage):
        """Calculate and apply damage to shield and hull."""

        if self._shield < damage:
            rest_damage = damage - self._shield
            self._shield = 0
            self._hull -= rest_damage
        else:
            self._shield -= damage
        if self._hull < 0:
            self._hull = 0

    def refresh_shield(self, amount=1):
        """Refresh shield."""

        if self._shield == self._max_shield:
            return

        if self._shield + amount > self._max_shield:
            self._shield = self._max_shield
        else:
            self._shield += amount


class GenericXEnemy(Ship):
    """Generic X enemy class."""

    def __init__(self, pos):
        super(GenericXEnemy, self).__init__(pos)
        # fmt: off
        self._image = Surface([
            "x^x",
            " X ",
            " * ",
        ])
        # fmt: on

        self._collider = Collider.simple(self)

        self.add_weapon(EBlaster())
        self._fire = True
        self._wbay = Point(x=self._image.width // 2, y=1)

        self._animgr = AnimationManager()

    # TODO: rethink this method
    #       Problem of emerging such methods is more complex problem
    #       of creating and configuring GameObjects.
    def add_animation(self, *args, **kwargs):
        self._animgr.add(*args, **kwargs)

    def update(self, dt):
        if self._hull <= 0:
            # TODO: [scoring]
            #       * Parametrize scores, move them to ships.conf
            #       * implement better scoring mechanism
            app.current().state.add_player_score(10)
            self.destroy()
            return

        self._animgr.update(dt)

        super(GenericXEnemy, self).update(dt)

    @collision.register("GenericXEnemy", "BasicPlasmaCannon")
    @collision.register("GenericXEnemy", "BasicLaserCharge")
    @collision.register("GenericXEnemy", "BasicUnguidedMissile")
    def collide(self, other, rect):
        self.take_damage(other.damage)
        other.destroy()


class PlayerShip(Ship):
    """PlayerShip class. Contains additional methods for HUD.

    :param :class:`xoinvader.utils.Point` pos: left top corner
    """

    def __init__(self, pos):
        super(PlayerShip, self).__init__(pos)

        # fmt: off
        self._image = Surface([
            "  O  ",
            "<=H=>",
            " * * ",
        ])
        # fmt: on

        self._pos = Point(
            x=pos.x - self._image.width // 2, y=pos.y - self._image.height
        )

        self._collider = Collider.simple(self)

        self._fire = False
        self._weapons = InfiniteList([Blaster(), Laser(), UM()])
        self._weapon = self._weapons.current()
        self._wbay = Point(x=self._image.width // 2, y=-1)

    def update_position(self, dt):
        """Update player ship position.

        Prohibits moving out behind the border.
        """

        border = Settings.layout.field.camera
        right = self._pos.x + self._image.width

        if right >= border.x - 1 and self._direction > 0:
            self._pos.x = border.x - self._image.width

        elif self._pos.x <= 1 and self._direction < 0:
            self._pos.x = 1

        else:
            # NOTE: Converting float to int reduces ship teleportation because
            #       we have no pixels.
            self._pos.x += int(self._direction * self._dx * dt / 1000)

        self._direction = 0

    def update(self, dt):
        if self._hull <= 0:
            app.current().trigger_state(
                "GameOverState", score=app.current().state.score
            )

        super(PlayerShip, self).update(dt)

    def get_weapon_info(self):
        """Return information about current weapon."""

        return "Weapon: {w} | [{c}/{m}]".format(
            w=self._weapon.type, c=self._weapon.ammo, m=self._weapon.max_ammo
        )

    @collision.register("PlayerShip", "EBasicPlasmaCannon")
    def collide(self, other, rect):
        self.take_damage(other.damage)
        other.destroy()



"""Module for common shared objects."""

import json
import pathlib
from os.path import dirname

import toml

import xoinvader
from xoinvader import constants
from xoinvader.utils import dotdict, Point


def get_json_config(path):
    """Return Settings object made from json.

    :param str path: path to config

    :return xoinvader.utils.dotdict: parsed config
    """

    with open(path) as fd:
        config = dotdict(json.load(fd))
    return config


def get_toml_config(path: str) -> dotdict:
    """Return Settings object made from toml."""

    with open(path) as fd:
        config = dotdict(toml.load(fd))
    return config


def get_config():
    """Return xoinvader configuration."""

    return get_toml_config(_CONFIG / "xoinvader.toml")


def rootify(root, config):
    """Append paths in dict to root.

    :param str root: common root
    :param dict config:

    :return dict: rootified config
    """

    for key, path in config.items():
        if isinstance(path, dict):
            rootify(root, path)
        else:
            config[key] = root / path
    return config


__all__ = ["Settings"]

WIDTH = 90
HEIGHT = 34

_ROOT = pathlib.Path(dirname(xoinvader.__file__))
_CONFIG = _ROOT / "config"

DEFAULT_XOI_SETTINGS = dict(
    layout=dict(
        field=dict(
            border=Point(x=WIDTH, y=HEIGHT),
            camera=Point(x=WIDTH - 1, y=HEIGHT - 1),
            player=Point(x=WIDTH // 2, y=HEIGHT - 1),
            edge=Point(x=WIDTH, y=HEIGHT - 1),
        ),
        gui=dict(
            bar=dict(
                health=Point(x=2, y=HEIGHT - 1),
                shield=Point(x=22, y=HEIGHT - 1),
                weapon=Point(x=WIDTH - 18, y=HEIGHT - 1),
            ),
            info=dict(weapon=Point(x=44, y=HEIGHT - 1)),
        ),
    ),
    path=rootify(_ROOT, get_json_config(_CONFIG / "path.json")),
    system=dict(debug=False, encoding=constants.UTF_8,),
)

Settings = dotdict(DEFAULT_XOI_SETTINGS)  # pylint: disable=invalid-name


def update_system_settings(args):
    """Update system settings from startup arguments.

    :param dict args: arguments
    """

    for arg, val in args.items():
        if arg in Settings.system:
            Settings.system[arg] = val
        else:
            raise KeyError("No such argument: '{0}'.".format(arg))



"""Module for creating and maintaining xoinvader levels.

One level consists of several files that can be packed into directory or ZIP file.

Without structure:
- levels
  - first-level.lv
  - first-level.bg
  - first-level.ev
  ...

Directory or zip archive
- levels
  - first-level.zip
  - second-level
    - second-level.lv
    - second-level.bg
    - second-level.ev

Usage:
level = Level.from_file("levels/first-level.zip")
level = Level.from_file("levels/second-level/second-level.lv")

state.add(level)
level.start()
"""


class LevelSyntaxError(Exception):
    def __init__(self, line, message):

        self.line = line
        self.message = message

        super().__init__(f"[line: {line}] {message}")


class LevelFormat:
    """Level file format.

    # Comments and whitespaces between sections are ignored
    # Comments cannot be placed within blocks
    # first meaningful line must be in format {engine}:{filespec}.
    xo1:level
    # or
    xo1:level@1

    If background chunks defined in separate file, you should specify it:
    bg:path/to/level-1.bg
    # TODO: inplace definition

    If events defined in separate file, you should specify it:
    ev:path/to/level-1.ev
    """

    ENGINE_XO1 = "xo1"
    SUPPORTED_ENGINES = [
        ENGINE_XO1,
    ]

    FORMAT_LEVEL = "level"
    """Generic level format."""

    FORMAT_LEVELv1 = "level@1"

    SUPPORTED_FORMATS = [
        FORMAT_LEVEL,
        FORMAT_LEVELv1,
    ]

    SEP = ":"
    COMMENT = "#"
    EVENT = f"ev{SEP}"
    BACKGROUND = f"bg{SEP}"

    def __init__(self):
        self._parsed = False
        self._current_block = None
        self._line = 0

        self._engine = None
        self._format = None

        self._bg_name = None
        self._bg_file = None

    def raise_syntax_error(self, message):
        raise LevelSyntaxError(self.line, message)

    def parse(self, filename):
        with open(filename) as level:
            self._file = level

            self.parse_magic()
            self._parsed = True

            print(self.as_string())
            # self.parse_background()
            # self.parse_events()
            # self.parse_level()

        # first meaningful line must contain format

    def _advance(self):
        if self._parsed:
            raise Exception("File already parsed")

        self.line += 1

        return self._file.readline().strip()

    def _split(self, line):
        return map(str.strip, line.split(self.SEP))

    def _splittable(self, line):
        return line and self.SEP in line

    def parse_magic(self):
        line = self._advance()
        if self._splittable(line):
            engine, filespec = self._split(line)
            if engine not in self.SUPPORTED_ENGINES:
                self.raise_syntax_error(f"engine '{engine}' is not supported")

            if filespec not in self.SUPPORTED_FORMATS:
                self.raise_syntax_error(f"format {filespec} is not supported")

            self._engine = engine
            self._format = filespec

    def parse_event(self, header):
        if not self._splittable(header):
            self.raise_syntax_error("can't parse event header")

        _, ev_name = self._split(header)
        if ev_name:
            self._ev_name = ev_name
            self._ev_file = open(self._ev_name)
        # else defined in level file

    def parse_background(self, header):
        if not self._splittable(header):
            self.raise_syntax_error("can't parse background header")

        _, bg_name = self._split(header)
        if bg_name:
            self._bg_name = bg_name
            self._bg_file = open(self._bg_name)
        # else defined in level file

    def as_string(self):
        if not self._parsed:
            raise Exception("File is not fully parsed")

        return f"""
        {self._engine}{self.SEP}{self._format}\n

        {self.EVENT}
        """


class Level(object):
    """Container for level event sequence and resources.

    Intended to use as just container, but may be subclassed. It's useful in
    order to add methods for creating concrete animations and events, such as
    spawning actual enemies, bonuses, initiating boss fights and so on.

    :param int speed: relative speed of the wave. Means how fast time advances.
    The faster time advances, the shorter the delays between events triggering.
    :param dict events: map from relative time in event to event starting
    function.
    :param bool running: if event sequence currently advances.
    """

    def __init__(self, speed=0):
        self._running = False
        self._counter = 0
        self._speed = speed
        self._events = {}
        self._event_timeouts = []

    @property
    def speed(self):
        """Relative wave's speed.

        :getter: yes
        :setter: yes
        :type: int
        """
        return self._speed

    @speed.setter
    def speed(self, value):
        """Setter."""
        self._speed = value

    @property
    def running(self):
        """If event sequence currently advances.

        :getter: yes
        :setter: no
        :type: bool
        """
        return self._running

    def add_event(self, time, callback):
        """Add event to some point in time.

        :param int time: point in time relative to level start when to run
        `callback`. Callback is fired when `_counter` exceeds provided value
        :param function callback: callback to be fired when wave reaches `time`
        """
        self._events.setdefault(time, []).append(callback)

    def start(self):
        """Start the level.

        Resets the counter, recomputes the list of event timeouts and sets the
        `running` property to `True`.
        """

        self._running = True
        self._counter = 0
        self._event_timeouts = sorted(self._events)

    def update(self):
        """Update the counter and fire appropriate events.

        Function is doing anything useful only when the Level has been started
        earlier, i.e. currently in running state.
        Function updates current timer and fires all events, which timeouts have
        expired. If there's no more events in the queue, the running state
        terminates and object goes to sleep.
        Time counter is increased by `_speed` value at each call of the
        fucntion.
        """

        if not self._running:
            return

        self._counter += self._speed
        while self._event_timeouts[0] <= self._counter:
            current_timeout = self._event_timeouts[0]
            for callback in self._events[current_timeout]:
                callback()
            del self._event_timeouts[0]
            if not self._event_timeouts:
                self._running = False
                break



"""InGameState-specific input/event handlers and commands."""

import curses
import logging
import weakref

from eaf.state import State

from xoinvader.background import Background
from xoinvader.collision import CollisionManager
from xoinvader.common import Settings
from xoinvader.style import Style
from xoinvader.gui import TextCallbackWidget, TextWidget, WeaponWidget, Bar
from xoinvader.handlers import EventHandler
from xoinvader.keys import KEY
from xoinvader.level import Level
from xoinvader.ship import GenericXEnemy, PlayerShip
from xoinvader.utils import Point, dotdict


LOG = logging.getLogger(__name__)


# pylint: disable=invalid-name
class TestLevel(Level):
    def __init__(self, state_object_adder, speed):
        super(TestLevel, self).__init__(speed)
        self._state_add = state_object_adder

        self._player_ship = PlayerShip(Settings.layout.field.player)
        self._state_add(self._player_ship)
        self._enemies = weakref.WeakSet()

        self.add_event(1, self.spawn4)
        self.add_event(100, lambda: None)
        self.add_event(200, self.del4)

        self.bg = Background(Settings.path.level1bg, speed=1, loop=True)
        self.bg.start(filled=True)
        self._state_add(self.bg)

    @staticmethod
    def get_keyframes(y_offset, x_start, direction):
        """Generate keyframe list.

        :param int y_offset: y of top left corner of image
        :param int x_start: x level of left corner of image
        :param int direction: direction of moving: -1 is left, 1 is right
        """

        def new_x(d):
            return x_start - d * direction

        return [
            (0, Point(x_start, y_offset)),
            (4.0, Point(x_start, y_offset + 20)),
            (7.0, Point(new_x(30), y_offset + 10)),
        ]

    def spawn4(self):
        right_side = Settings.layout.field.edge.x

        e1 = GenericXEnemy(Point(10, 1))
        e2 = GenericXEnemy(Point(25, 1))
        e3 = GenericXEnemy(Point(right_side - 10, 1))
        e4 = GenericXEnemy(Point(right_side - 25, 1))

        e1.add_animation(
            "", e1, "_pos", self.get_keyframes(1, 10, 1), interp=True
        )
        e2.add_animation(
            "", e2, "_pos", self.get_keyframes(1, 20, 1), interp=True
        )

        e3.add_animation(
            "",
            e3,
            "_pos",
            self.get_keyframes(1, right_side - 10, -1),
            interp=True,
        )
        e4.add_animation(
            "",
            e4,
            "_pos",
            self.get_keyframes(1, right_side - 20, -1),
            interp=True,
        )

        self._enemies = weakref.WeakSet([e1, e2, e3, e4])
        self._state_add(list(self._enemies))

    def del4(self):
        for enemy in self._enemies:
            enemy.destroy()


class InGameState(State):

    # TODO FIXME: collision manager looks like collision system
    #             belonging to State. Idea is to create some sort of
    #             mechanism for registering/deregistering/accesing
    #             theese systems via GameObject's current State.
    # ATTENTION: Here I met a problem of initialization of this state.
    #            1) here in init we created collision manager (now moved here)
    #            2) initialization of TestLevel created first game object
    #            (PlayerShip) that have collider
    #            3) Collider needs to be registered in CollisionManager that
    #            already was created, BUT! Collider.__init__ can't have
    #            access to current state because several frames above current
    #            control flow resides in State.__init__ and State does not
    #            fully instantiated yet and could not be accessed via
    #            application.state.
    #            One approach that I see here is to create optional postinit
    #            method and create objects that want access to state there.
    # NOTE: I think that here we must extrude systems with such behaviour and
    #       create them before state initialization, maybe here in classfields.
    collision = CollisionManager()

    def postinit(self):
        """Deferred initialization.

        Prepare GameObjects that require created and registered State object.
        """

        self.level = TestLevel(self.add, speed=1)
        # self.level.register_call("spawn_random", self.spawn_random)
        # self.level.register_call("spawn4", self.spawn4)

        self.actor = self.level._player_ship

        # TODO: [scoring]
        self.score = 0

        self._events = EventHandler(
            self,
            {
                KEY.A: self.actor.move_left,
                KEY.D: self.actor.move_right,
                KEY.E: self.actor.next_weapon,
                KEY.Q: self.actor.prev_weapon,
                KEY.R: lambda: self.actor.take_damage(5),
                KEY.SPACE: self.actor.toggle_fire,
                KEY.ESCAPE: self.pause_command,
            },
        )

        self.add(self._create_gui())
        self.level.start()

    def pause_command(self):
        self.app.state = "PauseMenuState"

    def add_player_score(self, amount):
        """Add player score.

        :param int amount: scores to add
        """

        self.score += amount

    def get_player_score_string(self):
        """Callback for TextCallbackWidget.

        :return str: score string
        """

        return "Score: {0}".format(self.score)

        # TODO: [collider-destruction]
        #       Remove this after collider instant destruction.
        #       Now colliders in weakref set garbage-collected with noticeable
        #       delay. In future we need to drop weakrefs and manually manage
        #       objects.
        LOG.debug("Colliders: %s", len(self.collision._colliders))
        LOG.debug("Collisions: %s", len(self.collision._collisions))
        LOG.debug("Objects in state: %s", len(self._objects))

    def _create_gui(self):
        """Create user interface."""

        comp = dotdict(
            whatever=lambda val: 0.0 <= val <= 100.0,
            high=lambda val: 70.0 <= val <= 100.0,
            normal=lambda val: 35.0 <= val < 70.0,
            low=lambda val: 0.0 <= val < 35.0,
        )

        return [
            Bar(
                pos=Settings.layout.gui.bar.health,
                prefix="Hull: ",
                general_style=curses.A_BOLD,
                stylemap={
                    comp.high: Style().gui["dp_ok"],
                    comp.normal: Style().gui["dp_middle"],
                    comp.low: Style().gui["dp_critical"],
                },
                callback=self.actor.get_hull_percentage,
            ),
            Bar(
                pos=Settings.layout.gui.bar.shield,
                prefix="Shield: ",
                general_style=curses.A_BOLD,
                stylemap={
                    comp.high: Style().gui["sh_ok"],
                    comp.normal: Style().gui["sh_mid"],
                    comp.low: Style().gui["dp_critical"],
                },
                callback=self.actor.get_shield_percentage,
            ),
            Bar(
                pos=Settings.layout.gui.bar.weapon,
                stylemap={comp.whatever: Style().gui["dp_ok"]},
                callback=self.actor.get_weapon_percentage,
            ),
            WeaponWidget(
                Settings.layout.gui.info.weapon, self.actor.get_weapon_info
            ),
        ] + [
            TextCallbackWidget(Point(2, 0), self.get_player_score_string),
            TextWidget(
                Point(Settings.layout.field.edge.x // 2 - 4, 0),
                "XOInvader",
                curses.A_BOLD,
            ),
        ]

    def events(self):
        self._events.handle()

    def update(self, dt):
        self.collision.update()
        self.level.update()
        if not self.level.running:
            self.level.start()
        LOG.info(self._objects)
        super().update(dt)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xoinvader', 'xoinvader.tests']

package_data = \
{'': ['*'],
 'xoinvader': ['config/*', 'res/*', 'res/gfx/*', 'res/tgfx/*'],
 'xoinvader.tests': ['fixtures/*']}

install_requires = \
['eaf>=0.2,<0.3', 'toml>=0.10,<0.11', 'xo1>=0.1,<0.2']

entry_points = \
{'console_scripts': ['xoigame = xoinvader.game:main']}

setup_kwargs = {
    'name': 'xoinvader',
    'version': '0.2.0',
    'description': '(not so) small python curses space game',
    'long_description': '|PyPI| |Build Status| |codecov.io|\n\n=========\nXOInvader\n=========\n\nAttempt to create small but serious game using **python** and\n**ncurses**.\n\nRequirements\n============\n\nGeneral\n-------\n\n* >= python-3.7\n* >= eaf-0.2\n* >= xo1-0.1\n* >= toml-0.10\n* ncurses >=5.9\n\nInstallation\n============\n\nFrom PyPI\n---------\n\n.. code-block:: console\n\n   $ pip install --user xoinvader\n\n\nDevelopment\n-----------\n\n.. code-block:: console\n\n   $ poetry install\n   $ poetry run xoigame\n\nTesting\n-------\n\n.. code-block:: console\n\n   $ make test  # for all tests\n   $ make view_cov  # run all tests and view coverage\n   $ # pass additional arguments to pytest via PYTEST_ARGS env variable\n   $ # you can pass args such way to test and view_cov make targets\n   $ make test PYTEST_ARGS=\'-m "not slow"\'  # don\'t run slow tests\n   $ pytest --markers  # show all markers\n   $ pytest xoinvader/path/to/test_file.py  # run separate file\n\n\nDocumentation\n-------------\n\n* Autogenerated from sources `here <http://xoinvader.mmap.me/>`__.\n* Development `wiki <https://github.com/pkulev/xoinvader/wiki/>`_.\n* Build documentation from sources:\n\n  .. code-block:: console\n\n     $ make docs\n     $ make view_docs  # build and open in firefox\n\n\n.. |PyPI| image:: https://badge.fury.io/py/xoinvader.svg\n   :target: https://badge.fury.io/py/xoinvader\n.. |Build Status| image:: https://travis-ci.org/pkulev/xoinvader.svg?branch=master\n   :target: https://travis-ci.org/pkulev/xoinvader\n.. |codecov.io| image:: http://codecov.io/github/pkulev/xoinvader/coverage.svg?branch=master\n   :target: http://codecov.io/github/pkulev/xoinvader?branch=master\n',
    'author': 'Pavel Kulyov',
    'author_email': 'kulyov.pavel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pkulev/xoinvader',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Test xoinvader.utils module."""

import pytest

from xoinvader.utils import (
    setup_logger,
    dotdict,
    isclose,
    InfiniteList,
    Point,
    Surface,
    Timer,
)


# pylint: disable=invalid-name,protected-access,missing-docstring
def test_setup_logger():
    logger = setup_logger("test", True)
    assert logger


def test_dotdict_setattr():
    settings = dotdict()
    settings.test_entry = 42
    assert settings["test_entry"] == 42
    assert settings.test_entry == 42

    settings["test_entry_2"] = 42
    assert settings.test_entry == 42

    assert pytest.raises(AttributeError, lambda: settings.bad_key)


@pytest.mark.parametrize(
    ("left", "right", "kwargs", "expected"),
    (
        (0.0, 0.0, {}, True),
        (1.0, 0.0, {}, False),
        (0.1, 0.0, {}, False),
        (0.1, 0.0, {"abs_tol": 0.1}, True),
        (0.1, 0.0, {"rel_tol": 0.1}, False),
    ),
)
def test_isclose(left, right, kwargs, expected):
    """xoinvader.utils.isclose"""
    assert isclose(left, right, **kwargs) is expected


def test_infinite_list_operations():

    # Test one element behaviour
    data = "test1"
    inf_list = InfiniteList([data])

    assert len(inf_list) == 1
    assert inf_list[0] == data
    assert inf_list.select(0) == data
    assert inf_list.current() == data
    assert inf_list.next() == data
    assert inf_list.prev() == data

    # TODO: Test many elements behaviour


def test_infinite_list_operations_negative():

    # Test empty InfiniteList behaviour
    inf_list = InfiniteList()
    for func in [
        inf_list.current,
        inf_list.next,
        inf_list.prev,
    ]:
        with pytest.raises(IndexError):
            func()

    # Negative tests for one element behaviour

    inf_list.append("test")

    with pytest.raises(IndexError):
        inf_list.select(-1)
    with pytest.raises(IndexError):
        inf_list.select(1)

    # TODO: test many elements behaviour


# fmt: off
_image = [
    [" ", "O", " "],
    ["x", "X", "x"],
]
# fmt: on


def test_surface_attributes():
    surface = Surface(_image)
    assert surface.height == len(_image)
    assert surface.width == len(_image[0])
    assert surface.raw == _image


def test_image_generator():
    surface = Surface(_image)
    image_gen = surface.get_image()
    for lpos, image, style in image_gen:
        assert _image[lpos.y][lpos.x] == image
        assert style is None


@pytest.mark.slow
def test_timer_get_elapsed():
    timer = Timer(5.0, lambda: True)
    timer.start()
    while timer.running:
        assert timer.get_elapsed() >= 0.0
        timer.update()
        assert timer.get_elapsed() >= 0.0



"""Keys' integer representation."""

from eaf.meta import Singleton


K_A = ord("a")
K_D = ord("d")
K_E = ord("e")
K_F = ord("f")
K_N = ord("n")
K_R = ord("r")
K_S = ord("s")
K_Q = ord("q")
K_W = ord("w")
K_SPACE = ord(" ")
K_ESCAPE = 27
K_ENTER = 343


class _KEY(metaclass=Singleton):
    """Convenient key container with dot access.

    K_KEYNAME can be accessed via KEY.KEYNAME.
    .. Attention:: this class relies to K_{KEYNAME} scheme.
                   No more than one underscore allowed.
    """

    KEY_PREFIX = "K_"
    """Prefix for all keys."""

    def __init__(self):
        for key, scancode in globals().items():
            if key.startswith(self.KEY_PREFIX):
                self.__dict__[key.split("_")[1]] = scancode


KEY = _KEY()
"""Key container object."""
